# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
from typing import Any
from typing import Iterable

import pydantic

from aiopki.utils import b64encode_json


class BaseJWT(pydantic.BaseModel):
    claims: dict[str, Any] = pydantic.Field(default={}, alias='__claims__')

    @pydantic.model_validator(mode='before')
    def preprocess(cls, values: bytes | str | dict[str, Any]) -> dict[str, Any]:
        if isinstance(values, (bytes, str)):
            values = json.loads(values)
        assert isinstance(values, dict)
        return {
            **values,
            '__claims__': {
                k: v
                for k, v in values.items()
                if k not in cls.model_fields
            }
        }
    
    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('exclude_defaults', True)
        kwargs.setdefault('exclude_unset', True)
        exclude = kwargs.setdefault('exclude', set())
        exclude.add('claims')
        return {**self.claims, **super().model_dump(**kwargs)}

    def encode(self, **kwargs: Any) -> bytes:
        """Encode the payload of the JWT."""
        return b64encode_json({
            **self.claims,
            **self.model_dump(**kwargs)
        })
    
    def has_claims(self, claims: Iterable[str]) -> bool:
        """Return a boolean indicating if the token has the given claims."""
        return set(claims) <= set(self.model_dump(exclude_none=True))


class JWT(BaseJWT):
    aud: str | list[str] | None = None
    iss: str | None = None
    exp: int | None = None
    sub: str | None = None

    def model_post_init(self, _: Any) -> None:
        if isinstance(self.aud, list) and len(self.aud) == 1:
            self.aud = self.aud[0]

    @pydantic.field_serializer('aud')
    def serialize_aud(self, _: Any) -> str | list[str] | None:
        aud = self.aud
        if isinstance(self.aud, list) and len(self.aud) == 1: # pragma: no cover
            aud = self.aud[0]
        return aud

    def has_audience(self, value: str) -> bool:
        return self.aud is not None and any([
            isinstance(self.aud, list) and value in self.aud,
            isinstance(self.aud, str) and value == self.aud
        ])


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from aiopki.types import Base64


class BaseEncryptionResult(pydantic.BaseModel):
    alg: str = pydantic.Field(
        default=...
    )

    kid: str | None = pydantic.Field(
        default=None
    )
    ct: Base64 = pydantic.Field(
        default=...
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Literal

import pydantic

import aiopki
from aiopki.ext import jose
from aiopki.types import IAlgorithm


class CryptoKeyVersion(aiopki.BaseCryptoKey):
    exp: int | None = None
    iat: int
    kty: Literal['oct', 'EC', 'RSA']
    name: str
    nbf: int | None = None
    jwk: jose.JWK
    uri: str

    @pydantic.field_validator('kty', mode='before')
    def preprocess_kty(cls, value: str) -> str:
        if value in {'EC-HSM', 'RSA-HSM'}:
            value = str.split(value, '-')[0]
        return value

    @pydantic.field_validator('exp', mode='before')
    def preprocess_exp(cls, value: int | datetime.datetime | None) -> int | None:
        if isinstance(value, datetime.datetime):
            value = int(value.timestamp())
        return value

    @pydantic.field_validator('iat', mode='before')
    def preprocess_iat(cls, value: int | datetime.datetime) -> int:
        if isinstance(value, datetime.datetime):
            value = int(value.timestamp())
        return value

    @pydantic.field_validator('nbf', mode='before')
    def preprocess_nbf(cls, value: int | datetime.datetime | None) -> int | None:
        if isinstance(value, datetime.datetime):
            value = int(value.timestamp())
        return value

    @property
    def alg(self) -> str:
        assert self.jwk.alg is not None
        return self.jwk.alg

    @property
    def public(self) -> jose.JWK | None:
        return self.jwk

    def default_algorithm(self) -> IAlgorithm:
        return aiopki.algorithms.get(self.alg)

    def is_available(self) -> bool:
        return self.jwk.is_available()

    def get_thumbprint(self) -> str:
        return self.jwk.thumbprint


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib
from typing import Any


class HMACSigning:
    __module__: str = 'aiopki.algorithms'
    name: str

    @property
    def digestmod(self) -> str:
        return self._digalg

    def __init__(self, digalg: str):
        self._digalg = digalg

    def get_digest_algorithm(self) -> Any:
        raise NotImplementedError

    def get_initialization_vector(self) -> Any:
        raise NotImplementedError

    def get_padding(self) -> Any:
        raise NotImplementedError

    def digest(self, value: bytes) -> bytes:
        return hashlib.new(self._digalg).digest()

    def supports_aad(self) -> bool:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from .ialgorithm import IAlgorithm


class ISigner(Protocol):
    __module__: str = 'aiopki.types'

    @property
    def thumbprint(self) -> str:
        ...

    def default_algorithm(self) -> IAlgorithm: ...
    async def sign(self, message: bytes, algorithm: IAlgorithm | None = None, using: str | None = None) -> bytes: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .algorithms import get as algorithm
from .algorithms import curve
from .algorithms import register
from .backends import install_backend
from .basecryptokey import BaseCryptoKey
from .cryptokeyspecification import CryptoKeySpecification
from .cryptokeyuri import CryptoKeyURI
from .secreturi import SecretURI
from . import types


__all__: list[str] = [
    'algorithm',
    'curve',
    'install_backend',
    'register',
    'types',
    'BaseCryptoKey',
    'CryptoKeySpecification',
    'CryptoKeyURI',
    'SecretURI',
]


def key(uri: str):
    pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

from .cryptokey import CryptoKey


__all__: list[str] = [
    'handles',
    'parse_key',
]


parse_key = CryptoKey.parse_uri


def handles(uri: urllib.parse.ParseResult) -> bool:
    return uri.scheme == 'file'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import re
import urllib.parse
from typing import cast
from typing import Iterable

import pydantic
from azure.identity.aio import DefaultAzureCredential
from azure.keyvault.keys.aio import KeyClient
from azure.keyvault.keys.crypto.aio import CryptographyClient

import aiopki
import aiopki.utils
from aiopki.ext import jose
from aiopki.types import IAlgorithm
from aiopki.types import ICryptoKey
from .cryptokeyversion import CryptoKeyVersion


credential: DefaultAzureCredential = DefaultAzureCredential()


EC_ALGORITHMS = {
    'P-384': 'ES384'
}


class CryptoKey(pydantic.BaseModel):
    default: str | None
    name: str
    vault_url: str
    _versions: list[CryptoKeyVersion] = []
    _index: dict[str, CryptoKeyVersion] = {}
    _loaded: bool = False


    @property
    def crv(self) -> str | None:
        assert self.default is not None
        return self._index[self.default].jwk.crv # type: ignore

    @property
    def thumbprint(self) -> str:
        assert self.default is not None
        return self._index[self.default].jwk.thumbprint

    @classmethod
    def parse_uri(cls, uri: urllib.parse.ParseResult):
        if uri.hostname is None: # pragma: no cover
            raise ValueError(f"Invalid URI: {uri}")
        if not re.match(r'^.*\.vault\.azure\.net', uri.hostname):
            raise ValueError("Not an Azure key or secret.")
        m = re.match(r'^/keys/([^/]+)(/[0-9a-z]+)?$', uri.path)
        if m is None:
            raise ValueError(f"Invalid URI: {uri.path}")
        name, version = m.groups()
        if version:
            version = re.sub(r'[^a-z0-9]+', '', version)
        return cls.model_validate({
            'name': name,
            'vault_url': f'{uri.scheme}://{uri.netloc}/',
            'default': version
        })

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return any([
            v.public.can_use(algorithm)
            for v in self._versions if v.public
        ])

    def default_algorithm(self) -> IAlgorithm:
        assert self.default is not None
        return self._index[self.default].default_algorithm()

    def versions(self) -> Iterable[ICryptoKey]:
        return self._versions

    async def discover(self, _client: KeyClient | None = None):
        now = int(datetime.datetime.now(datetime.timezone.utc).timestamp())
        if self._loaded: # pragma: no cover
            return self
        client: KeyClient | None = _client
        credential: DefaultAzureCredential | None = None
        if client is None:
            credential = DefaultAzureCredential()
            client = KeyClient(self.vault_url, credential=credential)
        try:
            versions: list[CryptoKeyVersion] = []
            async for props in client.list_properties_of_key_versions(self.name): # type: ignore
                assert props.created_on is not None
                if not props.enabled or props.version is None:
                    continue
                version = await client.get_key(props.name, version=props.version) # type: ignore
                public = None
                if str(version.key_type or '') in {'EC', 'EC-HSM', 'RSA', 'RSA-HSM'}:
                    kty = re.sub(r'^(oct|EC|RSA)(\-HSM)?$', r'\1', str(version.key_type))
                    public = {}
                    if version.key.crv: # type: ignore
                        public['alg'] = EC_ALGORITHMS[version.key.crv] # type: ignore
                    for attname in version.key._FIELDS: # type: ignore
                        v = getattr(version.key, attname)
                        if isinstance(v, bytes):
                            v = aiopki.utils.b64encode(v)
                        if attname == 'key_ops':
                            v = [str(x) for x in v]
                        public[attname] = v
                    public = jose.JWK.model_validate({
                        **public,
                        'kty': kty
                    })
                    assert public.alg is not None
                    versions.append(
                        CryptoKeyVersion.model_validate({
                            'name': props.version,
                            'exp': props.expires_on,
                            'kty': kty,
                            'iat': props.created_on,
                            'nbf': props.not_before,
                            'uri': props.id,
                            'jwk': public
                        })
                    )
                    self._index[public.thumbprint] = versions[-1]
        finally:
            if credential is not None:
                assert client is not None
                await credential.close() # type: ignore
                await client.close()

        # Set a default version that is not expired.
        self._versions = versions
        for version in self._versions:
            if version.exp is not None and version.exp <= now:
                continue
            if version.nbf is not None and version.nbf > now:
                continue
            self.default = version.jwk.thumbprint
            break
        if self.default not in self._index: # pragma: no cover
            raise ValueError(f"Default key {self.default} does not exist.")
        self._loaded = True
        return self

    async def sign(self, message: bytes, algorithm: IAlgorithm, using: str | None = None) -> bytes:
        if not self._loaded: # pragma: no cover
            raise RuntimeError(f"Call {type(self).__name__}.discover() or await object.")
        assert self.default is not None
        version = self._index[using or self.default]
        message = algorithm.digest(message)
        async with DefaultAzureCredential() as credential:
            async with CryptographyClient(version.uri, credential=credential) as client:
                client = cast(CryptographyClient, client)
                result = await client.sign(version.jwk.alg, message) # type: ignore
        return result.signature

    async def verify(
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm,
        using: str | None = None
    ) -> bool:
        assert self.default is not None
        version = self._index[using or self.default]
        assert version.public is not None
        return await version.public.verify(signature, message, algorithm, using)


    def __await__(self):
        return self.discover().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import importlib
import urllib.parse

from .types import ICryptoBackend


_BACKENDS: list[ICryptoBackend] = []


def install_backend(qualname: str, _version: str = 'v1') -> None:
    """Given the qualified name to a Python module, install
    the backend implementation of cryptographic operations.
    """
    module = importlib.import_module(qualname)
    if module not in _BACKENDS: # type: ignore
        _BACKENDS.insert(0, module) # type: ignore


def get_backend(uri: str | urllib.parse.ParseResult) -> ICryptoBackend:
    install_backend('aiopki.ext.cryptography')
    if not isinstance(uri, urllib.parse.ParseResult):
        uri = urllib.parse.urlparse(uri)
    backend = None
    for candidate in _BACKENDS:
        if not candidate.handles(uri):
            continue
        backend = candidate
        break
    if backend is None:
        raise ValueError("Unable to determine backend for URI.")
    return backend


def parse_uri(uri: str):
    backend = get_backend(uri)
    return backend.parse_key(urllib.parse.urlparse(uri))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any

import pytest
import pytest_asyncio

import aiopki
from aiopki.types import IAlgorithm
from aiopki.ext import azure


@pytest.fixture(scope='session')
def key_uri() -> urllib.parse.ParseResult:
    return urllib.parse.urlparse('https://python-oauthx.vault.azure.net/keys/ES256')


@pytest.fixture(scope='session')
def key_version_uri() -> urllib.parse.ParseResult:
    return urllib.parse.urlparse('https://python-oauthx.vault.azure.net/keys/ES256/2c19c50887fe49c7ae98acdd2022ab76')


@pytest.fixture(scope='session')
def vault_url() -> str:
    return 'https://python-oauthx.vault.azure.net/'


@pytest.fixture
def signing_algorithm() -> IAlgorithm:
    return aiopki.algorithms.get('ES256')


@pytest_asyncio.fixture(scope='session') # type: ignore
async def signing_key(key_uri: urllib.parse.ParseResult):
    k =  await azure.parse_uri(key_uri)
    assert k.default == '03510ac0fa574ab0aa539e82d41ee156'
    return k


@pytest_asyncio.fixture(scope='session') # type: ignore
async def verifier(signing_key: Any):
    return signing_key


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
import urllib.parse

from .cryptokey import CryptoKey
from .secret import Secret


__all__: list[str] = [
    'handles',
    'parse_key',
    'parse_secret'
]


def handles(uri: urllib.parse.ParseResult) -> bool:
    return uri.hostname is not None\
        and bool(re.match(r'^.*\.vault\.azure\.net', uri.hostname))


parse_key = CryptoKey.parse_uri
parse_secret = Secret.parse_uri